import {ConfMail} from "./types/conf";
import * as nodemailer from 'nodemailer'
import * as os from 'os'
import * as moment from 'moment'

const DEFAULT_FREQUENCY = 5 * 60 * 1000 // 5'
const queue: MailContent[] = []
let conf: ConfMail|undefined

export function initConfMail(confMail: ConfMail) {
  conf = confMail

  const freq = conf && conf.frequency || DEFAULT_FREQUENCY
  const freqS = freq / 1000
  console.log(`[conf] mails are sent every ${freqS.toFixed(0)}s`)
  ;(async () => {
    while (true) {
      await consumeQueue()
      await new Promise((res) => setTimeout(res, freq))
    }
  })()
}

export async function queueEmail(subject: string, html: string, cc?: string) {
  queue.push({ subject, body: html, cc })
  console.log(`[mail] added 1 mail to queue: ${subject}`)
}

function consumeMessages(cc: string): MailContent[] {
  const mails: MailContent[] = []
  for (let i = 0; i < queue.length; i++) {
    const mail = queue[i]
    if (mail.cc === (cc || undefined)) {
      mails.push(queue[i])
      queue.splice(i, 1)
      i--
    }
  }
  return mails
}

async function consumeQueue() {
  console.log(`[mail] consuming mailing queue: ${queue.length} messages to be sent`)
  while (queue.length) {
    if (queue.length > 1) {
      const ccs = queue.reduce((ccs, mail) => {
        const cc = getGroup(mail)
        if (ccs.indexOf(cc) === -1) {
          ccs.push(cc)
        }
        return ccs
      }, [] as string[])
      console.log(`[mail] grouping messages into ${ccs.length} group(s)`)
      for (const cc of ccs) {
        const messages: MailContent[] = consumeMessages(cc)
        const subject = `[dw] [${os.hostname}] ${messages.length} notifications`
        let body = ''
        for (const message of messages) {
          body += `<h3>${message.subject}</h3>\n`
          body += `<div>${message.body}</div>\n`
        }
        if (messages.length) {
          await sendEmail(subject, body, cc || undefined)
        }
      }
    } else {
      const mail = queue.shift() as MailContent
      await sendEmail(mail.subject, mail.body, mail.cc)
    }
  }
}

function getGroup(mail: MailContent) {
  return mail.cc || ''
}

async function sendEmail(subject: string, html: string, cc?: string) {

  if (!conf) {
    console.error(`Mail is not properly configured.`)
    return
  }

  console.debug(`[mail] Subject: ${subject}`)
  console.debug(`[mail] Body: ${html}`)
  if (!conf.enabled) {
    console.warn(`Mail is disabled.`)
    return
  }

  let transporter = nodemailer.createTransport({
    host: conf.host,
    port: conf.port,
    secure: false, // true for 465, false for other ports
    auth: {
      user: conf.username, // generated ethereal user
      pass: conf.apikey // generated ethereal password
    },
    authMethod: conf.auth,
    requireTLS: true,
  });

  // send mail with defined transport object
  let info = await transporter.sendMail({
    from: conf.from,
    to: conf.to,
    cc: (conf.cc ? [conf.cc] : []).concat(cc ? [cc] : []).join(','),
    subject,
    html
  })
}

export const mail = {

  onEstablished: (target: string,
                  message: () => string|undefined = () => `Connection established for ${target}`,
                  getHtml: () => string = () => `
      <p>
        Connection from [${os.hostname}] to ${target} established on ${moment().format('DD-MM-YYYY HH:mm:ss')}.
      </p>
    `) => {
    return async (cc?: string) => {
      await queueEmail(`[dw] [${os.hostname}] ${message()}`, getHtml(), cc)
      return undefined
    }
  },

  onDisconnect: (target: string,
                 message: () => string|undefined = () => `Connection closed for ${target}`,
                 getErrorMessage: () => string = () => '',
                 getHtml: (waitingDelay: number, recallDelay: number) => string = (waitingDelay: number, recallDelay: number) => `
      <p>
        Connection from [${os.hostname}] to ${target} was lost on ${moment().format('dd-MM-YYYY HH:mm:ss')}.
      </p>
      <p>
        Waiting ${(waitingDelay / 1000).toFixed(0)} seconds before trying to reconnect (${(recallDelay / 1000).toFixed(0)} seconds for recall).
      </p>
      ${getErrorMessage()}
  `) => {
    return async (waitingDelay: number, recallDelay: number, cc?: string) => {
      console.log('Waiting %s seconds...', (waitingDelay / 1000).toFixed(0))
      await queueEmail(`[dw] [${os.hostname}] ${message()}`, getHtml(waitingDelay, recallDelay), cc)
    }
  },

  onRestartSuccess: (target: string,
                     message: () => string|undefined = () => `Connection recovered for ${target}`,
                     getHtml: () => string = () => `
      <p>
        Connection from [${os.hostname}] to ${target} was recovered on ${moment().format('dd-MM-YYYY HH:mm:ss')}.
      </p>
    `) => {
    return async (cc?: string) => {
      console.log(`${message()}`)
      await queueEmail(`[dw] [${os.hostname}] ${message()}`, getHtml(), cc)
      return undefined
    }
  },
}

interface MailContent {
  subject: string
  body: string
  cc?: string
}