import {Conf, ConfHead} from "../../types/conf";
import {urlWatcher, UrlWatcherResult} from '../abstract/url-watcher'

function handleLateness(confHead: ConfHead, mainHeads: HeadMetric[], observedHead?: TrameWS2P): UrlWatcherResult {
  if (!mainHeads.length) {
    return UrlWatcherResult.ko(`Observed pubkey ${confHead.observedPubkey}: no consensus found`)
  }
  if (!observedHead) {
    return UrlWatcherResult.ko(`Observed pubkey ${confHead.observedPubkey} not found in heads`)
  }
  const matchingHeads = mainHeads.filter(h => h.blockstamp === observedHead.blockstamp)
  if (!matchingHeads.length) {
    // Check how much late is the node
    const farAwayHeads = mainHeads.filter(h => h.blockNumber - observedHead.blockNumber >= confHead.maxLateBlocks)
    if (farAwayHeads.length) {
      return UrlWatcherResult.ko(`Observed pubkey ${confHead.observedPubkey} is too late for ${farAwayHeads.length} consensus by at least ${confHead.maxLateBlocks}`)
    }
  }
  return UrlWatcherResult.ok()
}

export function headWatcher(conf: Conf) {

  const URL_PATH = '/network/ws2p/heads'
  const KEY_LENGTH = 10

  return async (confHead: ConfHead) => {

    return urlWatcher(conf, async (data) => {
      const heads = data as { heads: any[] }
      const mainHeads = getMain(heads)
      const observedHead = getObserved(heads, confHead.observedPubkey)
      return handleLateness(confHead, mainHeads, observedHead)
    })({
      name: `head ${confHead.address} => ${confHead.observedPubkey.substr(0, KEY_LENGTH)} (${confHead.name})`,
      address: confHead.address + URL_PATH,
      frequency: confHead.frequency
    },
        () => `State OK HEADS of ${confHead.observedPubkey.substr(0, KEY_LENGTH)}`,
        () => `State FAILURE HEADS of ${confHead.observedPubkey.substr(0, KEY_LENGTH)}`,
        () => `State RECOVERED HEADS of ${confHead.observedPubkey.substr(0, KEY_LENGTH)}`)
  }
}

function getMain(heads: { heads: any[] }): HeadMetric[] {
  const mapByHash: { [k: string]: HeadMetric } = {}
  for (const head of heads.heads) {
    const trame = getTrame(head)
    if (trame) {
      mapByHash[trame.blockstamp] = mapByHash[trame.blockstamp] || { pubkeys: [], blockNumber: trame.blockNumber, blockstamp: trame.blockstamp }
      mapByHash[trame.blockstamp].pubkeys.push(trame.pubkey)
    }
    else {
      console.warn('Pas de trame pour %s', head)
    }
  }
  return Object.keys(mapByHash).reduce((main, blockstamp) => {
    const newMetric = mapByHash[blockstamp]
    if (!main.length) {
      return [newMetric]
    }
    const newMain: HeadMetric[] = main.filter(m => m.pubkeys.length >= newMetric.pubkeys.length)
    let maxLen = main.length ? main[0].pubkeys.length : newMetric.pubkeys.length
    if (newMetric.pubkeys.length >= maxLen) {
      newMain.push(newMetric)
    }
    return newMain
  }, [] as HeadMetric[])
}

function getObserved(heads: { heads: any[] }, observedPubkey: string): TrameWS2P|undefined {
  const trame = heads.heads
      .map(getTrame)
      .filter(trame =>  trame && trame.pubkey === observedPubkey)[0]
  if (!trame) {
    return undefined
  }
  return trame
}

function getTrame(head: any): TrameWS2P|undefined {
  if (head.messageV2) {
    const [ protocol, type, trameVersion, pubkey, blockstamp, peerUuid, software, softVersion, n1, n2, n3 ] = head.messageV2.split(':')
    const [ blockNumber, blockHash ] = blockstamp.split('-')
    return { pubkey, blockstamp, blockNumber: parseInt(blockNumber), blockHash }
  }
  else if (head.message) {
    const [ protocol, type, trameVersion, pubkey, blockstamp, peerUuid, software, softVersion, n1 ] = head.message.split(':')
    const [ blockNumber, blockHash ] = blockstamp.split('-')
    return { pubkey, blockstamp, blockNumber: parseInt(blockNumber), blockHash }
  }
  else {
    return undefined
  }
}

interface HeadMetric {
  pubkeys: string[]
  blockstamp: string
  blockNumber: number
}

interface TrameWS2P {
  pubkey: string
  blockstamp: string
  blockNumber: number
  blockHash: string
}
